@@ -156,6 +156,18 @@ pub struct Stream {
 
     /// The last `PRIORITY_UPDATE` frame encoded field value, if any.
     last_priority_update: Option<Vec<u8>>,
+
+    /// The count of HEADERS frames that have been received.
+    headers_received_count: usize,
+
+    /// Whether a DATA frame has been received.
+    data_received: bool,
+
+    /// Whether a trailing HEADER field has been sent.
+    trailers_sent: bool,
+
+    /// Whether a trailing HEADER field has been received.
+    trailers_received: bool,
 }
 
 impl Stream {
@@ -196,6 +208,13 @@ impl Stream {
             data_event_triggered: false,
 
             last_priority_update: None,
+
+            headers_received_count: 0,
+
+            data_received: false,
+
+            trailers_sent: false,
+            trailers_received: false,
         }
     }
 
@@ -283,16 +302,39 @@ impl Stream {
             },
 
             Some(Type::Request) => {
-                // Request stream starts uninitialized and only HEADERS
-                // is accepted. Other frames cause an error.
+                // Request stream starts uninitialized and only HEADERS is
+                // accepted. After initialization, DATA and HEADERS frames may
+                // be acceptable, depending on the role and HTTP message phase.
+                //
+                // Receiving some other types of known frames on the request
+                // stream is always an error.
                 if !self.is_local {
                     match (ty, self.remote_initialized) {
-                        (frame::HEADERS_FRAME_TYPE_ID, false) =>
-                            self.remote_initialized = true,
+                        (frame::HEADERS_FRAME_TYPE_ID, false) => {
+                            self.remote_initialized = true;
+                        },
 
                         (frame::DATA_FRAME_TYPE_ID, false) =>
                             return Err(Error::FrameUnexpected),
 
+                        (frame::HEADERS_FRAME_TYPE_ID, true) => {
+                            if self.trailers_received {
+                                return Err(Error::FrameUnexpected);
+                            }
+
+                            if self.data_received {
+                                self.trailers_received = true;
+                            }
+                        },
+
+                        (frame::DATA_FRAME_TYPE_ID, true) => {
+                            if self.trailers_received {
+                                return Err(Error::FrameUnexpected);
+                            }
+
+                            self.data_received = true;
+                        },
+
                         (frame::CANCEL_PUSH_FRAME_TYPE_ID, _) =>
                             return Err(Error::FrameUnexpected),
 
@@ -400,7 +442,35 @@ impl Stream {
         let buf = &mut self.state_buf[self.state_off..self.state_len];
 
         let read = match conn.stream_recv(self.id, buf) {
-            Ok((len, _)) => len,
+            Ok((len, fin)) => {
+                // Check whether one of the critical stream was closed.
+                if fin &&
+                    matches!(
+                        self.ty,
+                        Some(Type::Control) |
+                            Some(Type::QpackEncoder) |
+                            Some(Type::QpackDecoder)
+                    )
+                {
+                    super::close_conn_critical_stream(conn)?;
+                }
+
+                len
+            },
+
+            Err(e @ crate::Error::StreamReset(_)) => {
+                // Check whether one of the critical stream was closed.
+                if matches!(
+                    self.ty,
+                    Some(Type::Control) |
+                        Some(Type::QpackEncoder) |
+                        Some(Type::QpackDecoder)
+                ) {
+                    super::close_conn_critical_stream(conn)?;
+                }
+
+                return Err(e.into());
+            },
 
             Err(e) => {
                 // The stream is not readable anymore, so re-arm the Data event.
@@ -440,6 +510,23 @@ impl Stream {
         self.local_initialized
     }
 
+    pub fn increment_headers_received(&mut self) {
+        self.headers_received_count =
+            self.headers_received_count.saturating_add(1);
+    }
+
+    pub fn headers_received_count(&self) -> usize {
+        self.headers_received_count
+    }
+
+    pub fn mark_trailers_sent(&mut self) {
+        self.trailers_sent = true;
+    }
+
+    pub fn trailers_sent(&self) -> bool {
+        self.trailers_sent
+    }
+
     /// Tries to fill the state buffer by reading data from the given cursor.
     ///
     /// This is intended to replace `try_fill_buffer()` in tests, in order to
@@ -1163,6 +1250,134 @@ mod tests {
         assert_eq!(stream.set_frame_type(frame_ty), Err(Error::FrameUnexpected));
     }
 
+    #[test]
+    fn additional_headers() {
+        let mut stream = Stream::new(0, false);
+
+        let mut d = vec![42; 128];
+        let mut b = octets::OctetsMut::with_slice(&mut d);
+
+        let header_block = vec![1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12];
+        let payload = vec![1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12];
+        let info_hdrs = frame::Frame::Headers {
+            header_block: header_block.clone(),
+        };
+        let non_info_hdrs = frame::Frame::Headers {
+            header_block: header_block.clone(),
+        };
+        let trailers = frame::Frame::Headers { header_block };
+        let data = frame::Frame::Data {
+            payload: payload.clone(),
+        };
+
+        info_hdrs.to_bytes(&mut b).unwrap();
+        non_info_hdrs.to_bytes(&mut b).unwrap();
+        data.to_bytes(&mut b).unwrap();
+        trailers.to_bytes(&mut b).unwrap();
+
+        let mut cursor = std::io::Cursor::new(d);
+
+        // Parse the HEADERS frame type.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_ty = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_ty, frame::HEADERS_FRAME_TYPE_ID);
+
+        stream.set_frame_type(frame_ty).unwrap();
+        assert_eq!(stream.state, State::FramePayloadLen);
+
+        // Parse the HEADERS frame payload length.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_payload_len = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_payload_len, 12);
+
+        stream.set_frame_payload_len(frame_payload_len).unwrap();
+        assert_eq!(stream.state, State::FramePayload);
+
+        // Parse the HEADERS frame.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        assert_eq!(stream.try_consume_frame(), Ok((info_hdrs, 12)));
+        assert_eq!(stream.state, State::FrameType);
+
+        // Parse the non-info HEADERS frame type.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_ty = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_ty, frame::HEADERS_FRAME_TYPE_ID);
+
+        stream.set_frame_type(frame_ty).unwrap();
+        assert_eq!(stream.state, State::FramePayloadLen);
+
+        // Parse the HEADERS frame payload length.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_payload_len = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_payload_len, 12);
+
+        stream.set_frame_payload_len(frame_payload_len).unwrap();
+        assert_eq!(stream.state, State::FramePayload);
+
+        // Parse the HEADERS frame.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        assert_eq!(stream.try_consume_frame(), Ok((non_info_hdrs, 12)));
+        assert_eq!(stream.state, State::FrameType);
+
+        // Parse the DATA frame type.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_ty = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_ty, frame::DATA_FRAME_TYPE_ID);
+
+        stream.set_frame_type(frame_ty).unwrap();
+        assert_eq!(stream.state, State::FramePayloadLen);
+
+        // Parse the DATA frame payload length.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_payload_len = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_payload_len, 12);
+
+        stream.set_frame_payload_len(frame_payload_len).unwrap();
+        assert_eq!(stream.state, State::Data);
+
+        // Parse the DATA payload.
+        let mut recv_buf = vec![0; payload.len()];
+        assert_eq!(
+            stream.try_consume_data_for_tests(&mut cursor, &mut recv_buf),
+            Ok(payload.len())
+        );
+        assert_eq!(payload, recv_buf);
+
+        assert_eq!(stream.state, State::FrameType);
+
+        // Parse the trailing HEADERS frame type.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_ty = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_ty, frame::HEADERS_FRAME_TYPE_ID);
+
+        stream.set_frame_type(frame_ty).unwrap();
+        assert_eq!(stream.state, State::FramePayloadLen);
+
+        // Parse the HEADERS frame payload length.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        let frame_payload_len = stream.try_consume_varint().unwrap();
+        assert_eq!(frame_payload_len, 12);
+
+        stream.set_frame_payload_len(frame_payload_len).unwrap();
+        assert_eq!(stream.state, State::FramePayload);
+
+        // Parse the HEADERS frame.
+        stream.try_fill_buffer_for_tests(&mut cursor).unwrap();
+
+        assert_eq!(stream.try_consume_frame(), Ok((trailers, 12)));
+        assert_eq!(stream.state, State::FrameType);
+    }
+
     #[test]
     fn zero_length_goaway() {
         let mut d = vec![42; 128];
